"""
Assumption calculation:
1) linearity
2) normality
3) homocedasticity
4) independence

"""


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import statsmodels.api as sm
from numpy.linalg import inv
from scipy import stats
import numpy as np

from statsmodels.compat import lzip
import statsmodels.stats.api as sms
from statsmodels.stats.stattools import durbin_watson
from .global_constants import GREEN, RED,  RESET

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


class Metrics:
    def __init__(self,data_x,data_y,significance_level,value_dl) -> None:
        self.data_x = data_x
        self.data_y = data_y
        self.alpha = significance_level
        self.value_dl = value_dl

        self.model_RLS = None
        self.standardized_residuals = None
        self.studentized_residuals = None
        self.enable_outlier = False
        self.x_one = None


        #run linear regression:
        self.linear_regression()
#-------------------------------------------------------------------------------
    def linear_regression(self):

        self.x_one = sm.add_constant(self.data_x)
        self.model_RLS=sm.OLS(self.data_y, self.x_one).fit()
        self.model_RLS.summary()
#--------------------------------------------------------------------------
    def linearity(self):
        value_linearity =  self.model_RLS.f_pvalue


        #1) validation:
        if value_linearity < self.alpha:
            print('1.2.1)'+GREEN+'Satisfies linearity'+RESET)
            enable_linearity = True
        else:
            print('1.2.1)'+RED+'Does not satisfies linearity'+RESET)
            enable_linearity = False


        return  enable_linearity
#-------------------------------------------------------------------------------
    def normality(self):


        #main:
        XTr=self.x_one.transpose()
        XM=np.dot(XTr,self.x_one )
        iXM=inv(XM)
        Mxi=np.dot(self.x_one , iXM)
        H=np.dot(Mxi,XTr)
        yPred=np.dot(H,self.data_y)
        Res=self.data_y-yPred
        #perform Shapiro-Wilk Test
        value_normality = stats.shapiro(Res).pvalue

        #1) validation:
        if value_normality > self.alpha:
            print('1.2.2)'+GREEN+ 'Satisfies normality'+RESET)
            enable_normality = True
        else:
            print('1.2.2)'+RED+ 'Does not satisfies normality'+RESET)
            enable_normality = False

        return enable_normality
#----------------------------------------------------------------------------------
    def homocedasticity(self):

        names=['Lagrange multiplier statistic', 'p-value','f-value', 'f p-value']
        test=sms.het_breuschpagan(self.model_RLS.resid, self.model_RLS.model.exog)
        values = lzip(names, test)

        value_homocedasticity = values[1][1]

        #1) validation:
        if value_homocedasticity > self.alpha:
            print('1.2.3)'+GREEN+'Satisfies homoscedasticity'+RESET)
            enable_homocedasticity = True
        else:
            print('1.2.3)'+RED+ 'Does not satisfies homoscedasticity'+RESET)
            enable_homocedasticity = False

        return enable_homocedasticity

#-----------------------------------------------------------------------------------------

    def independence(self):
        value_dL = self.value_dl

        #main:
        value_independence = durbin_watson(self.model_RLS.resid)

        #1) validation:
        if value_independence  > value_dL:
            print('1.2.4)'+GREEN+ 'Satisfies independence '+RESET)
            enable_independence  = True
        else:
            print('1.2.4)'+RED+'Does not satisfies independence '+RESET)
            enable_independence  = False

        return enable_independence

#-----------------------------------------------------------------------------------------
    def run(self):
        print('----'*20)
        #1) linearity:
        enable_line = self.linearity()
        #2) normality:
        if enable_line:
            enable_norm = self.normality()

            if enable_norm:
                enable_homo = self.homocedasticity()

                if enable_homo:
                    enable_inde = self.independence()
                    return enable_inde
                else:
                    return False

            else:
                 return False
        else:
            return False


"class for outlier detection"


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import statsmodels.api as sm
import numpy as np
import matplotlib.pyplot as plt 

from .global_constants import GREEN, RED,  RESET
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++





class OutlierDetection:

    def __init__(self,data_x,data_y) -> None:
        self.data_x = data_x
        self.data_y = data_y

        self.model_RLS = None
        self.standardized_residuals = None
        self.studentized_residuals = None
        self.enable_outlier = False

    def linear_regression(self):

        x_one = sm.add_constant(self.data_x)
        self.model_RLS=sm.OLS(self.data_y, x_one).fit()
        self.model_RLS.summary()

    def outlier_detection(self):

        #1) linear regression:
        self.linear_regression()

        #2) main:
        estandard =self.model_RLS.get_influence()
        self.standardized_residuals = estandard.resid_studentized_internal


        studentized_resid = self.model_RLS.outlier_test()
        self.studentized_residuals = studentized_resid[:,0]

        #3) validation
        size_data = self.standardized_residuals.shape[0]
        index_outlier = []
        limit_outlier = 3

        for idx_data in range(size_data ):

            if (np.abs(self.standardized_residuals[idx_data]) >= limit_outlier) and (np.abs(self.studentized_residuals[idx_data]) >=limit_outlier):
                index_outlier.append(idx_data)
                self.enable_outlier = True


        if self.enable_outlier:
            print('----'*20)
            print('Contains outlier: ')
            print(f'The following indices'+RED+ f'{index_outlier}'+RESET +'of the data x and y are deleted')

            new_x, new_y = self.elimination_outliers(index_outlier)

            return new_x, new_y
        else:
            print('----'*20)
            print(GREEN+'Does not contain outlier'+RESET)
            return self.data_x, self.data_y





    def elimination_outliers(self,idx_outlier):


        #1) elimination outliers:
        new_data_x = np.delete(self.data_x, idx_outlier, axis=0)
        new_data_y = np.delete(self.data_y, idx_outlier, axis=0)
        return new_data_x, new_data_y

    def run(self):

        x_final, y_final  = self.outlier_detection()

        return x_final, y_final


    def visualization(self):

        #run outlier detection:
        self.outlier_detection()


        #visualization
        plt.scatter(self.data_x,self.standardized_residuals)
        plt.xlabel('x', fontsize=18)
        plt.ylabel('Standardized Residuals', fontsize=18)
        plt.axhline(y=-2, color='black', linestyle='--', linewidth=1)
        plt.axhline(y=2, color='black', linestyle='--', linewidth=1)
        plt.axhline(y=0, color='black', linestyle='--', linewidth=1)
        plt.xticks(fontsize=14)
        plt.yticks(fontsize=14)
        plt.show()



        plt.scatter(self.data_x,self.studentized_residuals)
        plt.xlabel('x', fontsize=18)
        plt.ylabel('Studentized Residuals', fontsize=18)
        plt.axhline(y=0, color='black', linestyle='--', linewidth=1)
        plt.axhline(y=-2, color='black', linestyle='--', linewidth=1)
        plt.axhline(y=2, color='black', linestyle='--', linewidth=1)
        plt.xticks(fontsize=14)
        plt.yticks(fontsize=14)
        plt.show()


"""
Calculation of :
1) resolution
"""


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np
from scipy.stats import f_oneway
import copy
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def resolution(data_y,significance_level):

    #initialization:
    ySour = copy.deepcopy(data_y)
    alpha = significance_level
    Reso = None

    # Main Resolution: 
    y_mean = np.mean(ySour, axis=1)
    y_mean_new  = y_mean.reshape(ySour.shape[0], 1)
    

    sy = np.sort(y_mean_new, axis=None) # Order from smallest to largest
    dsy = np.diff(sy) # Diferncia consecutiva
    odsy = np.sort(dsy, axis=None) # dsy Order from smallest to largest


    i=0    
    for x in odsy:
        indexmin=np.where(dsy == odsy[i])[0]
        a=ySour[np.where(y_mean_new == sy[indexmin])[0],:].T.tolist()
        b=ySour[np.where(y_mean_new == sy[indexmin+1])[0],:].T.tolist()
        i=i+1
        if f_oneway(a, b).pvalue<alpha:
            Reso=odsy[i-1]            
            break

    return Reso


"""
Calls all functions and classes to generate the complete analysis:
1) calculation of assumptions
2) quality of the regression

"""


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
from art import text2art

from .outlier_detection import OutlierDetection
from .metrics import Metrics
from .parameters import Parameters
from .resolution import resolution
from .user_messages import parameters_messages
from .global_constants import BLUE,  RESET

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def regression_quality(x,y,alpha,dL,y_extended):
    space_1 = '            '
    space_2 = '                  '
    art_text1 = text2art("SIMPLE    LINEAR") 
    art_text2 = text2art(space_1+"REGRESION")  
    art_text3 = text2art(space_2+"QUALITY")  
   
    print(art_text1)
    print(art_text2)
    print(art_text3)
    #------------------------------------------------------------------------------------
    # 1) part:
    print(BLUE+'===='*20+RESET)
    print(BLUE+'part 1, Assumption calculation: '+RESET)
    print(BLUE+'===='*20+RESET)
    print('part 1.1, outlier identification: ')
    new_x, new_y = OutlierDetection(x,y).run()
    print('----'*20)

    # 2) part:
    print('part 1.2, Verification of: ')
    enable_metrics = Metrics(new_x,new_y,alpha,dL).run()

    # 3) part:
    print(BLUE+'===='*20+RESET)

    if enable_metrics:
        print(BLUE+'part 2, Regression quality: '+RESET)
        print(BLUE+'===='*20+RESET)
        #------------------------------------------
        accuracy,d_range,sensitivity = Parameters(new_x,new_y).run()
        #------------------------------------------
        res_value = resolution(y_extended,alpha)
        #-----------------------------------------
        parameters_messages(accuracy,res_value,sensitivity,d_range)
    else:
        print('Not all metrics were met')


from .main_routine import regression_quality
from .metrics import Metrics
from .outlier_detection import OutlierDetection
from .parameters import Parameters
from .resolution import resolution
from .user_messages import parameters_messages
from .global_constants import * 





# ANSI code for colors
RED = '\033[91m'
GREEN = '\033[92m'
YELLOW = '\033[93m'
BLUE = '\033[94m'
MAGENTA = '\033[95m'
CYAN = '\033[96m'
WHITE = '\033[97m'
RESET = '\033[0m'

#dielectric properties
DDX = 0.01 # spatial discretization
SPEED_C = 299792458 # speed of light [m/s]
DT = DDX/int(2*SPEED_C ) # temporal discretization
EPSZ = 8.8541e-12


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup (
    name='sl_regression_quality',
    version='0.1.3',
    license='MIT',
    description='simple linear regression quality',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author= 'EdelH, Aplata',
    packages=find_packages(),
    install_requires = ['numpy','pandas','statsmodels','scikit-learn','art'],

    url='https://github.com/aplatag/project_SL_regression_quality.git'
)



from .global_constants import CYAN,  RESET, BLUE


def parameters_messages(accuracy,resolution,sensitivity,range):
    
    print(f'2.1) Accuracy value:'+CYAN+ f'{round(accuracy,6)}'+RESET)
    print(f'2.2) Resolution value:' +CYAN+ f'{round(resolution,6)}'+RESET)
    print(f'2.3) Sensitivity value:' +CYAN+ f'{round(sensitivity,6)}'+RESET)
    print(f'2.4) Dynamic range:' +CYAN+ f'{round(range,6)}'+RESET)
    print(BLUE+'===='*20+RESET)


"""
calculation:

1) Accuracy
2) Range
3) Sensitivity
"""


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
from sklearn.model_selection import KFold
from sklearn.model_selection import cross_val_score
from sklearn.linear_model import LinearRegression
import statsmodels.api as sm
from numpy import mean
from numpy import absolute
import numpy as np

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class Parameters:

    def __init__(self,data_x,data_y) -> None:
        self.data_x = data_x
        self.data_y = data_y

        self.model_RLS = None
    #-------------------------------------------------------------------------------
    def linear_regression(self):

        self.x_one = sm.add_constant(self.data_x)
        self.model_RLS=sm.OLS(self.data_y, self.x_one).fit()
        self.model_RLS.summary()
    
    #-----------------------------------------------------------------------------------------------
    def accuracy(self):

        #define cross-validation method to use
        cv = KFold(n_splits=10, random_state=1, shuffle=True)

        #build multiple linear regression model
        model = LinearRegression()

        #use k-fold CV to evaluate model
        scores = cross_val_score(model, self.data_x, self.data_y , scoring='neg_root_mean_squared_error',
                                cv=cv, n_jobs=-1)

        value_accuracy = mean(absolute(scores))

        return value_accuracy
    #------------------------------------------------------------------------------------------------

    def dynamic_range(self):
         range_2 = np.max(self.data_y )
         range_1 = np.min(self.data_y)
         dyn_range = np.abs(range_2-range_1)
         return dyn_range
    #------------------------------------------------------------------------------------------------
    def sensitivity(self):
        self.linear_regression()
        sensi = np.abs(self.model_RLS.params[1])
        return sensi
    #------------------------------------------------------------------------------------------------
    def run(self):

        result_accuracy = self.accuracy()
        result_range = self.dynamic_range()
        result_sensitivity = self.sensitivity()

        return result_accuracy, result_range,result_sensitivity 
